import sys
import inspect
import argparse
import abc
import logging
import warnings

import argcomplete

import mini_buildd.misc
import mini_buildd.net
import mini_buildd.config

LOG = logging.getLogger(__name__)

# Early global log config (before the actual logging config steps in):
# - preliminary (stderr) handler
# - capture python warnings && setup 'py.warnings' logger
LOG_FORMAT = "%(levelname)s: %(message)s [%(name)s:%(lineno)d, thread=%(threadName)s]"
LOG_PRELIMINARY_HANDLER = logging.StreamHandler()
LOG_PRELIMINARY_HANDLER.setFormatter(logging.Formatter("mini-buildd preliminary console logger: " + LOG_FORMAT))
LOG = logging.getLogger("mini_buildd")
LOG.addHandler(LOG_PRELIMINARY_HANDLER)
LOGW = logging.getLogger("py.warnings")
LOGW.addHandler(LOG_PRELIMINARY_HANDLER)
logging.captureWarnings(True)


class CLI():
    def __init__(self, prog, description, epilog=None):
        self.prog = prog
        mini_buildd.misc.setup_console_logging(logging.DEBUG)

        self.parser = argparse.ArgumentParser(prog=prog, description=description, epilog=epilog,
                                              formatter_class=mini_buildd.misc.ArgumentDefaultsRawTextHelpFormatter)
        self.parser.add_argument("--version", action="version", version=mini_buildd.__version__)
        self.parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
                                 help="increase log level. Give twice for max logs")
        self.parser.add_argument("-q", "--quiet", dest="terseness", action="count", default=0,
                                 help="decrease log level. Give twice for min logs")
        self.args = None

    @classmethod
    def _add_endpoint(cls, parser):
        parser.add_argument("endpoint", action="store",
                            metavar="ENDPOINT",
                            help=f"HTTP target endpoint: {inspect.getdoc(mini_buildd.net.ClientEndpoint)}").completer = lambda **kwargs: mini_buildd.misc.DputCf().urls()

    @classmethod
    def _add_subparser(cls, subparser, cmd, doc):
        return subparser.add_parser(cmd, help=doc, formatter_class=mini_buildd.misc.ArgumentDefaultsRawTextHelpFormatter)

    def _loglevel(self):
        return logging.WARNING - (10 * (min(2, self.args.verbosity) - min(2, self.args.terseness)))

    @staticmethod
    def _daemon_log_handler_file():
        handler = logging.handlers.RotatingFileHandler(
            mini_buildd.config.LOG_FILE,
            maxBytes=5000000,
            backupCount=9,
            encoding="UTF-8")
        handler.setFormatter(logging.Formatter("%(asctime)s " + LOG_FORMAT))
        return handler

    @staticmethod
    def _daemon_log_handler_syslog():
        handler = logging.handlers.SysLogHandler(
            address="/dev/log",
            facility=logging.handlers.SysLogHandler.LOG_USER)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))
        return handler

    @staticmethod
    def _daemon_log_handler_console():
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(asctime)s " + LOG_FORMAT))
        return handler

    def _setup_daemon_logging(self, loggers):
        if self.args.foreground:
            loggers.append("console")

        # Clear all loggers now; this will remove the
        # preliminary console logger
        LOG.handlers = []
        LOGW.handlers = []

        # Try to add all loggers; collect exceptions to be able
        # to do error reporting later, when hopefully one valid
        # handler is set up.
        loggers_failed = {}
        for typ in loggers:
            try:
                handler_func = getattr(self, "_daemon_log_handler_" + typ)
                LOG.addHandler(handler_func())
                LOGW.addHandler(handler_func())
            except BaseException as e:
                loggers_failed[typ] = e

        # Set log level
        LOG.setLevel(self._loglevel())
        LOGW.setLevel(self._loglevel())

        # Properly set raiseException based "exception" debug value given (see https://docs.python.org/3.5/howto/logging.html#exceptions-raised-during-logging)
        logging.raiseExceptions = "exception" in mini_buildd.config.DEBUG

        if "warnings" in mini_buildd.config.DEBUG:
            warnings.simplefilter("default")

        # Finally, log all errors now that occurred while setting up loggers
        for typ, err in list(loggers_failed.items()):
            LOG.critical(f"Logger {typ} failed: {err}")

    def setup(self):
        pass

    @abc.abstractmethod
    def runcli(self):
        pass

    def run(self):
        argcomplete.autocomplete(self.parser)
        self.args = self.parser.parse_args()
        LOG.setLevel(self._loglevel())
        if LOG.getEffectiveLevel() <= logging.DEBUG:
            mini_buildd.config.DEBUG = ["exception"]

        try:
            self.setup()
            self.runcli()
        except BaseException as e:
            mini_buildd.config.log_exception(LOG, f"{self.prog} failed (try '-vv' to debug)", e)
            sys.exit(1)
