import enum
import queue
import datetime
import threading
import collections
import weakref
import logging


LOG = logging.getLogger(__name__)


class _NoLock():
    """
    Debug/temporary only: Experienced dead locks -- not reliably reproducible for now.

    May-be-related to our locking here. Use this to exclude our locking once we can
    reproduce the lockdown.
    """

    def acquire(self):
        pass

    def release(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, _type, _value, _traceback):
        pass


class Event(dict):
    Type = enum.Enum("Type", "BUILT,INSTALLED,FAILED,MIGRATED,REMOVED")

    def __init__(self, typ, package, version, distribution, extra=None, user=None, error=None):
        super().__init__()
        self["timestamp"] = str(datetime.datetime.now())
        self["type"] = typ.name
        self["package"] = package
        self["version"] = version
        self["distribution"] = distribution
        self["extra"] = extra
        self["user"] = user
        self["error"] = error


class Events(collections.deque):
    MAX_CLIENTS = 100

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._clients = {}
        self._lock = threading.Lock()

    def log(self, *args, **kwargs):
        event = Event(*args, **kwargs)
        LOG.debug(f"Logging event: {event['type']} {event['package']}")

        with self._lock:
            self.appendleft(event)

            for _key, (weakref_obj, q) in self._clients.items():
                if weakref_obj() is not None:
                    LOG.info(f"Queuing event for object '{weakref_obj}'")
                    q.put(event)

    def attach(self, obj):
        with self._lock:
            # Garbage collection
            for key in [key for key, (weakref_obj, _q) in self._clients.items() if weakref_obj() is None]:
                LOG.info(f"Detach event queue from object '{key}': Dead.")
                del self._clients[key]

            if id(obj) not in self._clients:
                if len(self._clients) >= Events.MAX_CLIENTS:
                    raise Exception(f"Events: Max client limit ({Events.MAX_CLIENTS}) exceeded")
                LOG.info(f"Attaching new event queue to object: {obj}")
                self._clients[id(obj)] = (weakref.ref(obj), queue.Queue())
                # Replay past events
                for event in self:
                    self._clients[id(obj)][1].put(event)

        return self._clients[id(obj)][1]

    def get_last_packages(self):
        return sorted({e["package"] for e in self if e["package"]})

    def get_last_versions(self, package):
        return sorted({e["version"] for e in self if e["package"] == package})
